/*
  fifo.h - First-In First-Out Buffer Template Library

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#include <string.h>
#include "task.h"

#define FIFO_NO_WAIT       0
#define FIFO_INFINITE_WAIT -1
#define FIFO_OVERWRITE     -2

template<typename T, size_t SIZE>
class Fifo : protected Sync
{
  private:

    size_t _wridx;
    size_t _rdidx;
    size_t _count;
    size_t _losts;
    T _buffer[SIZE];

    void read(T *buf, size_t count)
    {
      _count -= count;
      if (_rdidx + count > SIZE)
      {
        size_t n = SIZE - _rdidx;
        memcpy(buf, _buffer + _rdidx, n * sizeof(T));
        buf   += n;
        count -= n;
        _rdidx = 0;
      }
      memcpy(buf, _buffer + _rdidx, count * sizeof(T));
      _rdidx += count;
    }

    void write(const T *buf, size_t count)
    {
      _count += count;
      if (_wridx + count > SIZE)
      {
        size_t n = SIZE - _wridx;
        memcpy(_buffer + _wridx, buf, n * sizeof(T));
        buf   += n;
        count -= n;
        _wridx = 0;
      }
      memcpy(_buffer + _wridx, buf, count * sizeof(T));
      _wridx += count;
    }

  public:

    Fifo(void)
    : _wridx(0)
    , _rdidx(0)
    , _count(0)
    , _losts(0)
    {
    }

    size_t available(void)
    {
      size_t save = disableAndSaveInterrupts();
      size_t rv = _count;
      restoreInterrupts(save);
      return rv;
    }

    size_t availableForWrite(void)
    {
      return SIZE - available();
    }

    size_t losts(void)
    {
      size_t save = disableAndSaveInterrupts();
      size_t rv = _losts;
      _losts = 0;
      restoreInterrupts(save);
      return rv;
    }

    bool get(T *item, size_t count = 1, int32_t timeout = FIFO_INFINITE_WAIT)
    {
      size_t n, rv = 0;
      ctrl_t ctrl;
      while (count)
      {
        ctrl.alarm.interval(timeout);
        ctrl.retval = true;
        ctrl.save   = disableAndSaveInterrupts();
        if (_count)
        {
          n = count < _count ? count : _count;
          read(item, n);
          restoreInterrupts(ctrl.save);
          item  += n;
          count -= n;
          rv    += n;
        }
        else
        {
          if (timeout == FIFO_NO_WAIT)
            restoreInterrupts(ctrl.save);
          else if (sleep(ctrl))
            continue;
          break;
        }
      }
      return rv;
    }

    size_t put(const T *item, size_t count = 1, int32_t timeout = FIFO_INFINITE_WAIT)
    {
      size_t n, rv = 0;
      Alarm alarm;
      alarm.interval(timeout);
      while (count)
      {
        size_t save = disableAndSaveInterrupts();
        n = SIZE - _count;
        if (n)
        {
          if (count < n)
            n = count;
          write(item, n);
          restoreInterrupts(save);
          wakeup();
          item  += n;
          count -= n;
          rv    += n;
          alarm.interval(timeout);
        }
        else if (timeout == FIFO_OVERWRITE)
        {
          n = _count < count ? _count : count;
          _count -= n;
          _losts += n;
          if ((_rdidx += n) >= SIZE)
            _rdidx -= SIZE;
          restoreInterrupts(save);
        }
        else
        {
          restoreInterrupts(save);
          if (timeout == FIFO_NO_WAIT)
            break;
          else if ((timeout > FIFO_NO_WAIT) && alarm.expire())
            break;
          yield();
        }
      }
      return rv;
    }
};
